/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.geronimo;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Date;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.servlet.http.HttpServletResponse;

import org.apache.ahc.AsyncHttpClientCallback;
import org.apache.ahc.codec.HttpResponseMessage;

public class Callback implements AsyncHttpClientCallback {
    private final HttpServletResponse response;
    private final String sessionId;
    private final boolean debug;
    
    private final AtomicBoolean handled = new AtomicBoolean(false);
    
    private volatile Throwable throwable;
    private volatile HttpResponseMessage message;
    
    // in nanoseconds
    private volatile long time;

    public Callback(HttpServletResponse response, String sessionId, boolean debug) {
        this.response = response;
        this.sessionId = sessionId;
        this.debug = debug;
        clear();
    }
    
    /**
     * @param time
     */
    public void setTime(long time){
        this.time = time;
    }
    
    
    /* (non-Javadoc)
     * @see org.apache.ahc.AsyncHttpClientCallback#onResponse(org.apache.ahc.codec.HttpResponseMessage)
     */
    public synchronized void onResponse(HttpResponseMessage msg) {
        if (debug) {
            System.out.println("  onResponse()" + " - Status:" + msg.getStatusCode() + " - Length:" + msg.getContentLength() + " " + sessionId);
        }
        if (!handled.compareAndSet(false, true)) {
            System.out.println("already responded to another callback method " + sessionId);
            return;
        }

        synchronized(response) {
            response.setStatus(msg.getStatusCode());
            response.setContentType(msg.getContentType());
            if (msg.getContentLength() > 0) {
                response.setContentLength(msg.getContentLength());
                OutputStream out = null;
                try {
                    out = response.getOutputStream();
                    out.write(msg.getContent());
                    out.flush();
                    out.close();
                } catch (Exception e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
        if (debug) {
            System.out.println("  Responded to " + sessionId + " in " + (System.nanoTime() - this.time)/1000000);
        }
    }

    public synchronized void onException(Throwable cause) {
        throwable = cause;
        System.out.println("  Exception by " + sessionId + " in " + (System.nanoTime() - this.time)/1000000);       
        cause.printStackTrace();    
        if (!handled.compareAndSet(false, true)) {
            System.out.println("already responded to another callback method " + sessionId);
            return;
        }
    
        sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "exception occurred");
    }

    public synchronized void onClosed() {
        // if the response has not been received, this is an error: need to write a response
        if (!handled.get()) {
            sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Connection to the remote URL unexpectedly closed");
            System.out.println("Remote connection closed unexpectedly for " + sessionId);
        }
    }

    public synchronized void onTimeout() {
        System.out.println("  Timed out " + sessionId + " in " + (System.nanoTime() - this.time)/1000000);
        if (!handled.compareAndSet(false, true)) {
            System.out.println("already responded to another callback method " + sessionId);
            return;
        }
        
        sendError(HttpServletResponse.SC_REQUEST_TIMEOUT, "Request Timed out!\n" + new Date().toString());
    }
    
    private synchronized void sendError(int status, String msg) {
        synchronized(response) {
            try {
                if (response.isCommitted()) {
                    System.out.println("already responded to another callback method " + sessionId);
                    return;
                }
                response.setContentType("text/plain");
                response.setStatus(status);
                OutputStream out = response.getOutputStream();
                out.write(msg.getBytes());
                out.flush();
                out.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }


    public Throwable getThrowable() {
        return throwable;
    }

    public void clear() {
        message = null;
        throwable = null;
    }

    public HttpResponseMessage getMessage() {
        return message;
    }

    public void setMessage(HttpResponseMessage message) {
        this.message = message;
    }
}
